from gxl_ai_utils.utils import utils_file

input_list_path = "/home/work_nfs8/xlgeng/new_workspace/gxl_ai_utils/eggs/cats_and_dogs/prepare_data_for_salmonn_TTS/wenetspeech_acoustic/wtn_train.list"
utils_file.do_make_shard_file("")