from enum import Enum

# class ModelName(Enum):
#     FLOWER = 'ResNet18FlowerModel'
#     FLOWER_1 = 'ResNet152FlowerModel'
#     MNIST = 'MnistModel0'
#     GXLRNN = 'RNNGxlModel'
#     TEXTCLASSIFY_1 = 'CNNTextClassifyModel'
#     TEXTCLASSIFY_2 = 'RNNTextClassifyModel'

FLOWER = 'ResNet18FlowerModel'
FLOWER_1 = 'ResNet152FlowerModel'
MNIST = 'MnistModel0'
GXLRNN = 'RNNGxlModel'
TEXTCLASSIFY_1 = 'CNNTextClassifyModel'
TEXTCLASSIFY_2 = 'RNNTextClassifyModel'

GXL_SELF_ATTENTION = 'GxlAttention'
GXL_GPT_MLP = 'GPTMLP'
GXL_TRANSFORMER = 'GXLTransformer'