from gxl_ai_utils.utils import utils_file

if __name__ == "__main__":
    utils_file.download_file_by_request(
        "https://aod.cos.tx.xmcdn.com/storages/7b69-audiofreehighqps/8A/18/GKwRIMAJT0R7AFBzCgKLO2F9-aacv2-48K.m4a")
