# from . import store_data
# from . import store_model
from . import AiConstant
# from . import run as run
from . import utils as utils
from . import config as config
from . import thread as thread
# from . import gxl_dataset_wenet as gxl_dataset
# from . import audio_model as audio_model
# from . import audio_dataset as audio_dataset
# from . import audio_handler as audio_handler
# from . import model_warehouse as model_warehouse


def hello_ai():
    """write by gengxuelong"""
    print('hello gxl_ai_utils, here is xuelong geng ')
