import sys

import tqdm

from gxl_ai_utils.utils import utils_file



# 得到命令行参数
wav_scp_path = sys.argv[1]
text_scp_path = sys.argv[2]
output_dir = sys.argv[3]

utils_file.do_inference_paraformer(

)

