# from distutils.core import setup
from setuptools import setup, find_packages

setup(name='gxl_ai_utils',
      version='1.5.1',
      author='Xuelong Geng',
      description='这个是耿雪龙的工具包模块, update time: 2025-4-1',
      author_email='3349495429@qq.com',
      packages=find_packages(),)
